module spfab.fabricate;

private import std.c.windows.windows;
private import std.string;
private import std.math;

private import nscr.result;
private import nscr.charset;

private import spfab.vector2;
private import spfab.matrix;

// NScrLuaエクスポート関数
extern(C)
{
    bool NL_dostring (char *mem);
}

// NScripterエクスポート関数

struct DIBDATA {
    //NScripter内部のDIBのデータ。常に24bitで、DIBの通常の仕様どおりy座標が大きいほうから入っている。
    int width;
    int height;
    int color;
    byte *bits;
}

extern (C){
    typedef void function (int spnum, out DIBDATA d) GetSpriteDIBFuncPtr;
    typedef void function (int spnum, out int width, out int height, out int cellmax) GetSpriteDataFuncPtr;
}

private GetSpriteDIBFuncPtr p_GetSpriteDIB;
private GetSpriteDataFuncPtr p_GetSpriteData;

private void GetSpriteDIB(int spnum, out DIBDATA d){
    p_GetSpriteDIB(spnum, d);
}

private void GetSpriteData(int spnum, out int width, out int height, out int cellmax){
    p_GetSpriteData(spnum, width, height, cellmax);
}

void init(HMODULE hm){
    p_GetSpriteDIB = cast(GetSpriteDIBFuncPtr)GetProcAddress(hm, "GetSpriteDIB");
    p_GetSpriteData = cast(GetSpriteDataFuncPtr)GetProcAddress(hm, "GetSpriteData");
}

// プラグイン

enum GRAY {
    R = cast(int)(0.298912 * 1024),
    G = cast(int)(0.586611 * 1024),
    B = cast(int)(0.114478 * 1024)
}

align(1) struct RGBDATA {
    byte b;
    byte g;
    byte r;
    char[] toString() {
        return std.string.toString(r) ~ ":" ~ std.string.toString(g) ~ ":" ~ std.string.toString(b);
    }
    int average(){
        return ((r&0xff) + (g&0xff) + (b&0xff))/3;
    }
    uint grayscale(){
        return (GRAY.R*(r&0xff) + GRAY.G*(g&0xff) + GRAY.B*(b&0xff)) >> 10;
    }
    void set(byte r, byte g, byte b){
        this.r = r;
        this.g = g;
        this.b = b;
    }
    void set(int r, int g, int b){
        this.r = toByte(r);
        this.g = toByte(g);
        this.b = toByte(b);
    }
    void add(int dr, int dg, int db){
        this.r = toByte((r & 0xff) + dr);
        this.g = toByte((g & 0xff) + dg);
        this.b = toByte((b & 0xff) + db);
    }
    bool isTransparent(){
        if((this.r & 0xff) != 0xff) return false;
        if((this.g & 0xff) != 0xff) return false;
        if((this.b & 0xff) != 0xff) return false;
        return true;
    }
    bool isOpaque(){
        return !isTransparent();
    }
    int asAlpha(){
        // 計算用に逆にする。高速化のためRのみ採用。
        return (0xff - (r & 0xff));
    }
}

/// スプライトを表現する。
class Sprite {
    enum Alpha{
        TRANSPARENT = -1,
        OPAQUE = 0
    }
    
    /// スプライトを構築する。
    /**
        @param spnum スプライト番号
    **/
    this(int spnum){
        this.spnum = spnum;
        GetSpriteDIB(spnum, dib);
        GetSpriteData(spnum, width, height, cell);
    }
    
    /// a指定画像かどうか確認する。
    /**
        @retval true a指定
        @retval false a以外の指定
    **/
    bool hasAlpha(){
        return dib.width == width * 2 * cell;
    }
    
    /// 画素ごとに処理する。
    /**
        @param processPixcel 画素ごとの処理
    **/
    void processByPixcel(void delegate(inout RGBDATA pix) processPixcel){
        RGBDATA* rgb = cast(RGBDATA*)dib.bits;
        int w = width;
        int h = height;
        int ox = 0;
        int withmask = hasAlpha() ? 2 : 1;
        int scanlen = dib.width;
        for(int cell = this.cell - 1; cell >= 0; cell--){
            ox = cell * w * withmask;
            for(int y = h - 1; y >= 0; y--){
                for(int x = ox + w - 1; x >= ox; x--){
                    processPixcel(rgb[y * scanlen + x]);
                }
            }
        }
    }
    
    
    /// 指定セルのマスクについて画素ごとに処理する。
    /**
        マスクが無いスプライトでは呼んではならない。
        @param cell セル番号
        @param processPixcel 画素ごとの処理
    **/
    void processMaskInCellByPixcel(int cell, void delegate(inout RGBDATA pix) processPixcel){
        assert(hasAlpha());
        RGBDATA* rgb = cast(RGBDATA*)dib.bits;
        int w = width;
        int h = height;
        int ox = cell * w * 2 + w;
        int scanlen = dib.width;
        for(int y = h - 1; y >= 0; y--){
            for(int x = ox + w - 1; x >= ox; x--){
                processPixcel(rgb[y * scanlen + x]);
            }
        }
    }


    /// DIBデータを複製する。
    /**
        @return DIBデータ
    **/
    DIBDATA duplicateDIB(){
        DIBDATA dib = this.dib;
        byte[] bits = new byte[dib.width * 3 * dib.height];
        bits[0..$] = this.dib.bits[0..bits.length];
        dib.bits = cast(byte*)bits;
        return dib;
    }
    
    /// アルファ値でマスクを満たす。
    /**
        @param alpha アルファ値
    **/
    void fillAlpha(byte alpha){
        if(!hasAlpha()) return;
        RGBDATA* rgb = cast(RGBDATA*)dib.bits;
        RGBDATA alrgb = RGBDATA(alpha, alpha, alpha);
        for(int cell = this.cell - 1; cell >= 0; cell--){
            int ox = cell * width * 2 + width;
            for(int y = height - 1; y >= 0; y--){
                for(int x = ox + width - 1; x >= ox; x--){
                    rgb[y * dib.width + x] = alrgb;
                }
            }
        }
    }
    
    void clear(){
        dib.bits[0..dib.width-1] = 0;
        fillAlpha(Alpha.TRANSPARENT);
    }

    /// 画像を変換する。
    /**
        @param dstWidth 変換後の幅
        @param dstHeight 変換後の高さ
        @param matrix 変換行列(オフセット調整を除く)
        @param autoresize true:自動サイズ調整、false:調整なし
    **/
    void transform(int dstWidth, int dstHeight, ref Matrix23 matrix, bool autoresize){
        dstWidth += dstWidth % 2 == 0 ? 0 : 1; // 奇数幅だとラスタライズがおかしい・・・
        // 上書きするためデータを複製して保持
        DIBDATA srcDIB = duplicateDIB();
        RGBDATA* srcRGB = cast(RGBDATA*)srcDIB.bits;
        RGBDATA* dstRGB;
        int dstScanLength;
        int withmask = hasAlpha() ? 2 : 1;
        if(autoresize){
            // 転送先を作成
            NL_dostring(cast(char*)format(
                "local n = %d; local x,y,a = NSSpGetPos(n); NSSpLoad(n,\":%s/%d;>%d,%d,#000000\"); NSSpMove(n,x,y,a)", 
                spnum, hasAlpha() ? "a" : "c", cell, dstWidth * cell * withmask, dstHeight));
            auto dstSprite = new Sprite(spnum);
            dstSprite.fillAlpha(-1);
            dstRGB = cast(RGBDATA*)dstSprite.dib.bits;
            dstScanLength = dstSprite.dib.width;
        }
        else{
            // 元画像をクリアして利用
            clear();
            dstRGB = cast(RGBDATA*)this.dib.bits;
            dstScanLength = this.dib.width;
        }
        int srcWidth = this.width;
        int srcHeight = this.height;
        int srcScanLength = this.dib.width;
        version(DEBUG) char[] buf = "";
        // 変換行列を作成
        Vector2 offset = Vector2(dstWidth/2-0.5, dstHeight/2-0.5);
        Matrix23 preMtx, postMtx;
        preMtx.setTranslation(-offset);
        postMtx.setTranslation(offset);
        Matrix23 compMtx = postMtx;
        compMtx *= matrix;
        compMtx *= preMtx;
        // 4点作成
        Vector2 a, b, c, d;
        compMtx.multiply(a, Vector2((dstWidth - srcWidth)/2, (dstHeight - srcHeight)/2)); // LT
        compMtx.multiply(b, Vector2((dstWidth + srcWidth)/2 - 1, (dstHeight - srcHeight)/2)); // RT
        compMtx.multiply(c, Vector2((dstWidth - srcWidth)/2, (dstHeight + srcHeight)/2 - 1)); // LB
        d = a + (b - a) + (c - a);
        version(DEBUG) buf ~= a.toString() ~ " " ~ b.toString() ~ " " ~ c.toString() ~ " " ~ d.toString();
        // 転送元画像側
        Vector2 ai = Vector2(0, 0);
        Vector2 bi = Vector2(srcWidth - 1, 0);
        Vector2 ci = Vector2(0, srcHeight - 1);
        Vector2 di = Vector2(bi.x, ci.y);
        version(DEBUG) MessageBoxA(null, cast(char*)(buf ~ "\0"), cast(char*)("\0"), MB_OK);
        // 計算量としてはコピーの方が少ないので、1回のラスタライズで各セル・マスク全てを網羅する。
        int wcount = cell * withmask;
        Vector2 dstMin, dstMax;
        int gapX, gapY;
        if(autoresize){
            dstMin = Vector2(0, 0);
            dstMax = Vector2(dstWidth - 1, dstHeight - 1);
            gapX = 0;
            gapY = 0;
        }
        else{
            dstMin = Vector2((dstWidth - srcWidth)/2, (dstHeight - srcHeight)/2);
            dstMax = Vector2((dstWidth + srcWidth)/2 - 1, (dstHeight + srcHeight)/2 - 1);
            gapX = -(dstWidth - srcWidth)/2;
            gapY = -(dstHeight - srcHeight)/2;
            dstWidth = autoresize ? dstWidth : srcWidth; // リサイズが無い場合は元の幅分のメモリしか確保してない
        }
        rasterize(a, b, c, ai, bi, ci,
            delegate void (int ix, int iy, int x, int y){
                for(int windex = wcount - 1; windex >= 0; windex--){
                    if(0 <= ix && ix < srcWidth && 0 <= iy && iy < srcHeight)
                        dstRGB[(y+gapY) * dstScanLength + windex * dstWidth + (x+gapX)] = srcRGB[iy * srcScanLength + windex * srcWidth + ix];
                }
            }, dstMin, dstMax);
        rasterize(d, b, c, di, bi, ci,
            delegate void (int ix, int iy, int x, int y){
                for(int windex = wcount - 1; windex >= 0; windex--){
                    if(0 <= ix && ix < srcWidth && 0 <= iy && iy < srcHeight)
                        dstRGB[(y+gapY) * dstScanLength + windex * dstWidth + (x+gapX)] = srcRGB[iy * srcScanLength + windex * srcWidth + ix];
                }
            }, dstMin, dstMax);
    }
    
    void blend(Sprite another, int offsetX, int offsetY, void delegate(inout RGBDATA pix, RGBDATA pix2, int alpha, int alpha2) processPixcel){
        RGBDATA* rgb = cast(RGBDATA*)dib.bits;
        RGBDATA* rgb2 = cast(RGBDATA*)another.dib.bits;
        int maxX = another.width - 1;
        int maxY = another.height - 1;
        int w2 = another.width;
        int w = width;
        int h = height;
        int ox = 0;
        int withmask = hasAlpha() ? 2 : 1;
        int scanlen = dib.width;
        int ox2 = 0;
        int scanlen2 = another.dib.width;
        for(int cell = this.cell - 1; cell >= 0; cell--){
            ox = cell * w * withmask;
            ox2 = cell * w2 * withmask;
            for(int y = maxY; y >= 0; y--){
                for(int x = maxX; x >= 0; x--){
                    int index = (y + offsetY) * scanlen + (ox + x + offsetX);
                    int index2 = y * scanlen2 + x + ox2;
                    processPixcel(rgb[index], rgb2[index2], rgb[index + w].asAlpha(), rgb2[index2 + w2].asAlpha());
                    //processPixcel(rgb[index], rgb2[index2], 0xff, 0xff);
                }
            }
        }
    }
    
    version(DEBUG){
        void dump(){
            RGBDATA* rgb = cast(RGBDATA*)dib.bits;
            char[] buf;
            for(int cell = this.cell - 1; cell >= 0; cell--){
                int ox = cell * width * 2 + width;
                for(int y = height - 1; y >= 0; y--){
                    for(int x = ox + width - 1; x >= ox; x--){
                        buf ~= rgb[y * dib.width + x].toString();
                    }
                    buf ~= "\r\n";
                }
            }
            MessageBoxA(null, cast(char*)(buf ~ "\0"), cast(char*)("\0"), MB_OK);
        }
    }
    
    DIBDATA dib;
    int width;
    int height;
    int cell;
    int spnum;
}


/// スプライト画像を閾値で白黒に塗り分ける。
/**
    @param spnum 対象のスプライト番号
    @param threshold 閾値(閾値>RGB平均値なら黒,閾値<=RGB平均値なら白)
    @param lowwhite trueの場合、閾値>RGB平均値の場合を白とする。
    @retval false 失敗
    @retval true  成功
**/
bool monocro(int spnum, int threshold, bool lowwhite){
    auto sprite = new Sprite(spnum);
    // 各セルについて実行する。
    RGBDATA hi, lo;
    if(lowwhite){
        hi.set(0, 0, 0);
        lo.set(-1, -1, -1);
    }
    else{
        hi.set(-1, -1, -1);
        lo.set(0, 0, 0);
    }
    sprite.processByPixcel(delegate (inout RGBDATA pix){
        if(threshold > pix.average())
            pix.set(hi.r, hi.g, hi.b);
        else
            pix.set(lo.r, lo.g, lo.b);
        }
    );
    return true;
}


/// スプライト画像を白黒256階調にする。
/**
    @param spnum 対象のスプライト番号
    @retval false 失敗
    @retval true  成功
**/
bool grayscale(int spnum){
    auto sprite = new Sprite(spnum);
    // 各セルについて実行する。
    sprite.processByPixcel(delegate (inout RGBDATA pix){
            byte gray = cast(byte)(pix.grayscale());
            pix.set(gray, gray, gray);
        }
    );
    return true;
}


/// スプライト画像を中心で水平に反転する。
/**
    @param spnum 対象のスプライト番号
    @retval false 失敗
    @retval true  成功
**/
bool flipX(int spnum){
    auto sprite = new Sprite(spnum);
    // 各セルについて実行する。
    RGBDATA* rgb = cast(RGBDATA*)sprite.dib.bits;
    int w = sprite.width;
    int h = sprite.height;
    int ox = 0;
    int withmask = sprite.hasAlpha() ? 2 : 1;
    int scanlen = sprite.dib.width;
    for(int cell = sprite.cell * withmask - 1; cell >= 0; cell--){
        ox = cell * w;
        for(int y = h - 1; y >= 0; y--){
            for(int x = (w - 1)/2; x >= 0; x--){
                swap(rgb[y * scanlen + ox + x], rgb[y * scanlen + ox + (w - 1 - x)]);
            }
        }
    }
    return true;
}


/// スプライト画像を中心で垂直に反転する。
/**
    @param spnum 対象のスプライト番号
    @retval false 失敗
    @retval true  成功
**/
bool flipY(int spnum){
    auto sprite = new Sprite(spnum);
    // 各セルについて実行する。
    RGBDATA* rgb = cast(RGBDATA*)sprite.dib.bits;
    int w = sprite.width;
    int h = sprite.height;
    int ox = 0;
    int withmask = sprite.hasAlpha() ? 2 : 1;
    int scanlen = sprite.dib.width;
    for(int cell = sprite.cell * withmask - 1; cell >= 0; cell--){
        ox = cell * w;
        for(int y = (h - 1)/2; y >= 0; y--){
            for(int x = w - 1; x >= 0; x--){
                swap(rgb[y * scanlen + ox + x], rgb[(h - 1 - y) * scanlen + ox + x]);
            }
        }
    }
    return true;
}


/// スプライト画像を拡大・縮小する。
/**
    @param spnum 対象のスプライト番号
    @param zx 水平方向の拡大縮小率(1.0:そのまま)
    @param zy 垂直方向の拡大縮小率(1.0:そのまま)
    @param autoresize true:自動サイズ調整、false:調整なし
    @retval false 失敗
    @retval true  成功
**/
Result zoom(int spnum, double zx, double zy, bool autoresize){
    assert(zx > 0.0);
    assert(zy > 0.0);
    auto sprite = new Sprite(spnum);
    int w = round(sprite.width * zx);
    int h = round(sprite.height * zy);
    Matrix23 matrix;
    matrix.setScaling(Vector2(zx, zy));
    sprite.transform(w, h, matrix, autoresize);
    return Result(1, "");
}


/// スプライト画像を回転する。
/**
    @param spnum 対象のスプライト番号
    @param degree 回転度
    @param autoresize true:自動サイズ調整、false:調整なし
    @retval false 失敗
    @retval true  成功
**/
bool rotate(int spnum, int degree, bool autoresize){
    auto sprite = new Sprite(spnum);
    // 回転時にはみ出る大きさ(対角線-長辺)分拡大させて確保する。
    double diagonal = sqrt(cast(double)(sprite.width * sprite.width + sprite.height * sprite.height));
    int w = round(diagonal);
    int h = w;
    Matrix23 matrix;
    matrix.setRotation(degree);
    sprite.transform(w, h, matrix, autoresize);
    return true;
}


/// スプライト画像に色を加減算する。
/**
    @param spnum 対象のスプライト番号
    @param dr 加減算する赤成分
    @param dg 加減算する緑成分
    @param db 加減算する青成分
    @retval false 失敗
    @retval true  成功
**/
bool colorBy(int spnum, int dr, int dg, int db){
    auto sprite = new Sprite(spnum);
    sprite.processByPixcel(delegate (inout RGBDATA pix){
        pix.add(dr, dg, db);
    });
    return true;
}


/// スプライトをセル単位で画像・マスクが交互となるように配置しなおす。
/**
    PNGでのマルチセル用と思ったが、nspng.dllを使う場合は必要なかった。落涙。
    @param spnum 対象のスプライト番号
    @retval false 失敗
    @retval true  成功
**/
bool alternate(int spnum){
    auto sprite = new Sprite(spnum);
    if(sprite.cell < 2) return true;
    if(!sprite.hasAlpha()) return true;
    // 各セルについて実行する。
    byte* bits = sprite.dib.bits;
    int ox;
    int x1, x2;
    int w = sprite.dib.width * RGBDATA.sizeof;
    int h = sprite.height;
    int swap_width;
    byte[] buffer = new byte[(w-sprite.width*2)*RGBDATA.sizeof]; // 最大で(セル数-1)*セル幅を使う。
    // マージソートの要領でセル数-1..1の交換を行う。
    for(int cell = sprite.cell - 1; cell > 0; cell--){
        ox += sprite.width * RGBDATA.sizeof; // n回目の始点はセル幅
        swap_width = sprite.width * cell * RGBDATA.sizeof; // n回目の交換範囲は(セル数-回数)*セル幅*マスクで、交換するのでその半分
        for(int y = h - 1; y >= 0; y--){
            x1 = y * w + ox;
            x2 = x1 + swap_width;
            buffer[0..swap_width-1] = bits[x1..x2-1];
            bits[x1..x2-1] = bits[x2..x2+swap_width-1];
            bits[x2..x2+swap_width-1] = buffer[0..swap_width-1];
        }
    }
    return true;
}


/// 指定した座標がスプライトの不透過部分かどうかを確認する。
/**
    PNGでのマルチセル用と思ったが、nspng.dllを使う場合は必要なかった。落涙。
    @param spnum 対象のスプライト番号
    @param x x座標(スプライト左上を原点として相対指定)
    @param y y座標(スプライト左上を原点として相対指定)
    @param cell 対象のセル番号
    @retval false 失敗
    @retval true  成功
**/
bool ptinopaque(int spnum, int x, int y, int cell){
    auto sprite = new Sprite(spnum);
    if(cell >= sprite.cell){
        throw new Exception("セル番号が不正です。セル数-1以下を指定してください。");
    }
    if(x < 0 || sprite.width < x){
        return false;
    }
    if(y < 0 || sprite.height < y){
        return false;
    }
    RGBDATA* rgb = cast(RGBDATA*)sprite.dib.bits;
    int w = sprite.width;
    int ox = cell * w * 2 + w;
    int scanlen = sprite.dib.width;
    return rgb[y * scanlen + ox + x].isOpaque();
}


/// 指定したスプライトを透過値を見て合成する。
/**
**/
bool blend(int spnum, int blendnum, int x, int y){
    auto sprite = new Sprite(spnum);
    auto blendsp = new Sprite(blendnum);
    // セル数が一致していないと合成できない。
    if(sprite.cell != blendsp.cell) return false;
    // 座標がサイズを超えていると合成できない。
    if(sprite.width < blendsp.width + x) return false;
    if(sprite.height < blendsp.height + y) return false;
    // 透過が一致していないと合成できない。
    if(!sprite.hasAlpha() || !blendsp.hasAlpha()) return false;
    sprite.blend(blendsp, x, y, delegate (inout RGBDATA pix, RGBDATA pix2, int alpha, int alpha2){
        alpha = 0xff - alpha2;
        pix.set((toInt(pix.r) * alpha + toInt(pix2.r) * alpha2) / 0xff, (toInt(pix.g) * alpha + toInt(pix2.g) * alpha2) / 0xff, (toInt(pix.b) * alpha + toInt(pix2.b) * alpha2) / 0xff);
    });
    return true;
}


/// 指定したスプライトを加算合成する。
/**
**/
bool blendAdd(int spnum, int blendnum, int x, int y){
    auto sprite = new Sprite(spnum);
    auto blendsp = new Sprite(blendnum);
    // セル数が一致していないと合成できない。
    if(sprite.cell != blendsp.cell) return false;
    // 座標がサイズを超えていると合成できない。
    if(sprite.width < blendsp.width + x) return false;
    if(sprite.height < blendsp.height + y) return false;
    // 透過が一致していないと合成できない。
    if(!sprite.hasAlpha() || !blendsp.hasAlpha()) return false;
    sprite.blend(blendsp, x, y, delegate (inout RGBDATA pix, RGBDATA pix2, int alpha, int alpha2){
        pix.set((toInt(pix.r) * alpha + toInt(pix2.r) * alpha2) / 0xff, (toInt(pix.g) * alpha + toInt(pix2.g) * alpha2) / 0xff, (toInt(pix.b) * alpha + toInt(pix2.b) * alpha2) / 0xff);
    });
    return true;
}


/// 指定したスプライトを減算合成する。
/**
**/
bool blendSub(int spnum, int blendnum, int x, int y){
    auto sprite = new Sprite(spnum);
    auto blendsp = new Sprite(blendnum);
    // セル数が一致していないと合成できない。
    if(sprite.cell != blendsp.cell) return false;
    // 座標がサイズを超えていると合成できない。
    if(sprite.width < blendsp.width + x) return false;
    if(sprite.height < blendsp.height + y) return false;
    // 透過が一致していないと合成できない。
    if(!sprite.hasAlpha() || !blendsp.hasAlpha()) return false;
    sprite.blend(blendsp, x, y, delegate (inout RGBDATA pix, RGBDATA pix2, int alpha, int alpha2){
        pix.set((toInt(pix.r) * alpha - toInt(pix2.r) * alpha2) / 0xff, (toInt(pix.g) * alpha - toInt(pix2.g) * alpha2) / 0xff, (toInt(pix.b) * alpha - toInt(pix2.b) * alpha2) / 0xff);
    });
    return true;
}


private:
void swap(inout RGBDATA pix1, inout RGBDATA pix2){
    RGBDATA tmp = pix1;
    pix1 = pix2;
    pix2 = tmp;
}


byte toByte(int value){
    if(value <= 0)
        return 0;
    return cast(byte)(value > 0xff ? 0xff : value);
}


int toInt(byte value){
    return (value & 0xff);
}


int round(double value){
    value += (value > 0.0) ? 0.5 : -0.5;
    return cast(int)value;
}


void rotateVector(out Vector2 dst, ref Vector2 src, ref Vector2 offset, ref Matrix22 matrix){
    Vector2 tmpOffset = Vector2(-0.5, -0.5);
    tmpOffset += offset;
    dst.set(src - tmpOffset);
    matrix.multiply(dst, dst);
    dst += tmpOffset;
}


void rasterize(ref Vector2 a, ref Vector2 b, ref Vector2 c, 
    ref Vector2 ai, ref Vector2 bi, ref Vector2 ci, 
    void delegate(int, int, int, int) copyPixProc,
    ref Vector2 dstMin, ref Vector2 dstMax){
    Vector2 p0, p1, p2;
    Vector2 i0, i1, i2;
    if(a.y < b.y){
        if(b.y < c.y){ //abc
            p0 = a;  p1 = b;  p2 = c;
            i0 = ai; i1 = bi; i2 = ci;
        }
        else if(a.y < c.y){//acb
            p0 = a;  p1 = c;  p2 = b;
            i0 = ai; i1 = ci; i2 = bi;
        }
        else{//cab
            p0 = c;  p1 = a;  p2 = b;
            i0 = ci; i1 = ai; i2 = bi;
        }
    }
    else{
        if(a.y < c.y){ //bac
            p0 = b;  p1 = a;  p2 = c;
            i0 = bi; i1 = ai; i2 = ci;
        }
        else if(b.y < c.y){//bca
            p0 = b;  p1 = c;  p2 = a;
            i0 = bi; i1 = ci; i2 = ai;
        }
        else{//cba
            p0 = c;  p1 = b;  p2 = a;
            i0 = ci; i1 = bi; i2 = ai;
        }
    }
    // p0p2と分割水平線の交点座標p3を求める。p1上で分割するので、yはp1.y
    Vector2 p01 = p1 - p0;
    Vector2 p02 = p2 - p0;
    double t = p01.y / p02.y; // 辺p0p1とp0p2の比は色々使えるらしい。
    Vector2 p3 = Vector2(p0.x + p02.x * t, p1.y);
    Vector2 p03 = p3 - p0;
    // 元画像側の添字と先画像上の位置関係を式にする。
    Vector2 i01 = i1 - i0;
    Vector2 i02 = i2 - i0;
    double a0, a1, a2, a3, a4, a5;
    a0 = ( i01.x * p02.y - i02.x * p01.y ) / ( p01.x * p02.y - p02.x * p01.y );
    a1 = ( i01.x * p02.x - i02.x * p01.x ) / ( p01.y * p02.x - p02.y * p01.x );
    a2 = i0.x - p0.x * a0 - p0.y * a1;
    a3 = ( i01.y * p02.y - i02.y * p01.y ) / ( p01.x * p02.y - p02.x * p01.y );
    a4 = ( i01.y * p02.x - i02.y * p01.x ) / ( p01.y * p02.x - p02.y * p01.x );
    a5 = i0.y - p0.x * a3 - p0.y * a4;
    assert(p01.y == p03.y);
    // 上半分の三角形p0p1p3について処理
    if(p1.y > p0.y + 1.0){
        int yBeg = cast(int)fmax(p0.y, dstMin.y);
        int yEnd = cast(int)fmin(p1.y, dstMax.y);
        double rcpP01y = 1.0 / p01.y;
        version(DEBUG) char[] buf;
        for(int y = yBeg; y <= yEnd; ++y){
            double yf = y;
            // 辺p1p0、p3p0と今のyの水平線の交点を求める。
            t = (y - p0.y) * rcpP01y;
            double x0 = p0.x + p01.x * t;
            double x1 = p0.x + p03.x * t;
            int xBeg, xEnd;
            if(x0 > x1){
                xBeg = cast(int)fmax(x1, dstMin.x);
                xEnd = cast(int)fmin(x0, dstMax.x);
            }
            else{
                xBeg = cast(int)fmax(x0, dstMin.x);
                xEnd = cast(int)fmin(x1, dstMax.x);
            }
            for(int x = xBeg; x <= xEnd; ++x){
                double xf = x;
                int ix = cast(int)round(a0 * xf + a1 * yf + a2);
                int iy = cast(int)round(a3 * xf + a4 * yf + a5);
                // copy
                copyPixProc(ix, iy, x, y);
            }
            version(DEBUG) buf ~= toString(xBeg) ~ "," ~ toString(xEnd) ~ " ";
        }
        version(DEBUG) MessageBoxA(null, cast(char*)(buf ~ "\0"), cast(char*)("\0"), MB_OK);
    }
    // 下半分の三角形p1p2p3について処理する。
    if(p2.y > p1.y + 1.0){
        Vector2 p21 = p1 - p2;
        Vector2 p23 = p3 - p2;
        int yBeg = cast(int)fmax(p1.y, dstMin.y);
        int yEnd = cast(int)fmin(p2.y, dstMax.y);
        double rcpP21y = 1.0 / p21.y;
        for(int y = yBeg; y <= yEnd; ++y){
            double yf = y;
            // 辺p1p2、p3p2と今のyの水平線の交点を求める。
            t = (y - p2.y) * rcpP21y;
            double x0 = p2.x + p21.x * t;
            double x1 = p2.x + p23.x * t;
            int xBeg, xEnd;
            if(x0 > x1){
                xBeg = cast(int)fmax(x1, dstMin.x);
                xEnd = cast(int)fmin(x0, dstMax.x);
            }
            else{
                xBeg = cast(int)fmax(x0, dstMin.x);
                xEnd = cast(int)fmin(x1, dstMax.x);
            }
            for(int x = xBeg; x <= xEnd; ++x){
                double xf = x;
                int ix = cast(int)round(a0 * xf + a1 * yf + a2);
                int iy = cast(int)round(a3 * xf + a4 * yf + a5);
                // copy
                copyPixProc(ix, iy, x, y);
            }
        }
    }
}
